from __future__ import unicode_literals

__all__ = (
    'COMMAND',
    'PROMPT',
)


#: Name of the command buffer.
COMMAND = 'COMMAND'

#: Name of the input for a "command-prompt" command.
PROMPT = 'PROMPT'
